# print formeated html to stdout
package eGuideDog::Browser::HTML::Format;

use strict;
use IO::File;
use HTML::TreeBuilder;
use CGI;
use eGuideDog::Browser::HTML::Navigate;

sub new;
sub traverse;
sub printFormatedHtml;
sub filter_page;
sub reformat_page;
sub rule_extract_links;
sub rule_append_links;
sub split_page;
sub filter_text;

#my @links;
#our $browser; # LWP client browser user agent

sub new {
  my $self = {};

  bless $self, __PACKAGE__;
  return $self;
}

# print formated html
# input: $url
# output: print html code of the $url to STDOUT
sub printFormatedHtml($) {
  my $url = shift;

  # get file and build html tree
  my ($html, $tree);
  $html = get($url);
  defined $html || die "$! $url";
  $tree = HTML::TreeBuilder->new_from_content($html);

  # traverse and print the html tree
  traverse($tree, 0);
}

# traverse the tree
# input: $root -- root of the html DOM tree
#        $depth -- depth of the current, default is 0
# output: print the DOM tree in plain text to STDOUT
my $lineNo = 0;
sub traverse {
  my $root = shift;
  return -1 unless defined $root;
  bless($root, 'HTML::Element');
  my $depth = shift;
  $depth = 0 unless defined $depth;
  if ($depth > 100) {
    warn "traverse depth greater than 100.";
    return -1;
  }

#  print "tag: \'", $root->tag, "\'\n";
  print "  " x $depth, $root->starttag, "\n";

  my @nodes = @{$root->content_array_ref};
  foreach my $node (@nodes) {
    if (ref $node) {
      traverse($node, $depth + 1);
    } else {
      print "  " x ($depth + 1);
      print "$node\n";
    }
  }

  print "  " x $depth, $root->endtag, "\n";

  return 0;
}

# reformat html page
# input: an html DOM tree
# output: an html DOM tree
sub reformat_page($) {
  my $tree = shift;
  bless $tree, "HTML::Element";

  # move all links to the end of the page
  # it seems that this is not a good filter
#  rule_extract_links($tree);
#  rule_append_links($tree);

  return $tree;
}

# filter html page
# input: an html DOM tree
# output: an html DOM tree
sub filter_page {
  my ($self, $tree) = @_;

  ####### rules to filter the html tree ######

  my $begin = time;

  # rules relative to encoding
  #rule_encoding($tree);
  #debug('rule_encoding: ' . (time - $begin));
  #$begin = time;

  # rule of deleting invisible elements
  rule_delete_invisible($tree);
#  debug('rule_delete_invisible: ' . (time - $begin));
  $begin = time;

  # rule of translating elements cannot be spoken to elements can be spoken
  rule_trans_invoicable($tree);
#  debug('rule_trans_invoicale: ' . (time - $begin));
  $begin = time;

  # rule of re-constructure the document
  rule_restruct($tree); # cost too much time
#  debug('rule_restruct: ' . (time - $begin));
  $begin = time;

  # divide document into several parts and build structure summary
  rule_struct_summary($tree);
#  debug('rule_struct_summary: ' . (time - $begin));
  $begin = time;

  # find main part, in 'rule_struct_summary.pl'
  rule_main_part($tree);
#  debug('rule_main_part: ' . (time - $begin));
  $begin = time;

  # add item number 'e-guidedog_no'
#  rule_add_number($tree); # this is done in buffer.pl

  return $tree;
}

# extract all links
# precondition: @links
sub rule_extract_links($) {
  my $node = shift;
  bless $node, "HTML::Element";

  my @links;

  my @child_nodes = @{$node->content_array_ref};
  foreach my $child_node (@child_nodes) {
    if (ref $child_node) {
      # <a href="address"...>description</a>
      if ($child_node->tag eq "a") {
	if (defined $child_node->attr("href")) {
	push @links, $child_node;
	$child_node->detach();
      }
      } else {
	rule_extract_links($child_node);
      }
    }
  }
}

# append extracted links to the end of the page
# precondition: @links
# postcondition: add links to the end of the page
sub rule_append_links($) {
  my $root = shift;
  bless $root, "HTML::Element";

  my @links;

#  my $node = ${$root->content_array_ref}[0];
  my @child_nodes = @{$root->content_array_ref};
  foreach my $child_node (@child_nodes) {
    if (ref $child_node and $child_node->tag eq "body") {
      my $count = @links;
      for (my $i = 0; $i < $count; $i++) {
	$child_node->push_content(@links);
      }
    }
  }
}

sub split_page($) {
  my $root = shift;
  bless($root, 'HTML::Element');

  foreach my $child_node ($root->content_refs_list) {
    my $child_node = $$child_node;
    if (ref $child_node) {
      if ($child_node->tag eq "map") {
	split_map($child_node);
      }
    }
  }
}

sub split_map($) {
# input a map node
  my $node_map = shift;
  bless $node_map, "HTML::Element";

  my @links;

  foreach my $node_area ($node_map->content_refs_list) {
    my $node_area = $$node_area;
    if (ref $node_area) {
      if ($node_area->tag eq "area") {
	my $link = '<a href="' . $node_area->attr('href') . '" value=' . $node_area->attr('href') . '></a>';
	push(@links, $link);
      } else {
	warn "something wrong!";
      }
    } else {
      warn "something wrong!";
    }
  }

  # generate new page for the map
  my $map_path = generate_tmpfile();
  open(FILE, '>', $map_path);
  print FILE '<html><body>';
  print FILE @links;
  print FILE '</body></html>';
  close(FILE);

  # substitude img including usemap
  my $root = $node_map->root;
  my @node_imgs = $root->find_by_attribute('usemap', '#' . $node_map->attr('name'));
  foreach my $node_img (@node_imgs) {
    if ($node_img->tag eq 'img') {
	my $link = HTML::Element->new('a', 'href'=>"file://$map_path", 'value'=>'site map ' . $node_map->attr('name'));
	$node_img->preinsert($link);
	$node_img->delete();
    } else {
      warn 'encount a usemap in <' . $node_img->tag . '>!';
    }
  }

  $node_map->delete();
}

# show text will be navigated
sub filter_text($) {
  my $node = shift;
  bless $node, "HTML::Element";

  foreach my $child_node ($node->content_refs_list) {
    my $child_node = $$child_node;
    if (ref $child_node) {
      # delete <map>
      if ($child_node->tag eq 'map') {
	$child_node->delete();
      }

      # delete <option>
      if ($child_node->tag eq 'option') {
	$child_node->delete();
      }

      # recursively check the child nodes
      else {
	filter_text($child_node);
      }
    }
  }

  return $node;
}

# ------ rule_delete_invisible.pl ----- #
sub rule_delete_invisible($);
sub rule_delete_blank_lines($);
sub rule_delete_invisible_element($);

sub rule_delete_invisible($) {
  my $node = shift;
  bless $node, "HTML::Element";

  rule_delete_blank_lines($node);

  foreach my $child_node ($node->content_refs_list) {
    $child_node = $$child_node;
    if (ref $child_node) {
      rule_delete_invisible_element($child_node);
    }
  }
}

sub rule_delete_blank_lines($) {
  my $node = shift;
  bless $node, 'HTML::Element';

  my @child_nodes = $node->detach_content();
  foreach my $child_node (@child_nodes) {
    if (ref $child_node) {
      $node->push_content($child_node);
      rule_delete_blank_lines($child_node);
    } else {
      # filter blank line
#      $child_node =~ s/^\s+//;
#      $child_node =~ s/\s+$//;
#      if ($child_node =~ /\w/
#	  || $child_node =~ /^\/$/) { # reserve '/' for link to root
      $child_node =~ /^\s*(.*)\s*$/;
      if ($1) {
	$node->push_content($child_node);
      }
#      }
    }
  }
}

sub rule_delete_invisible_element($) {
  my $child_node = shift;
  bless $child_node, 'HTML::Element';

      # delete <style>
      if ($child_node->tag eq 'style') {
	$child_node->delete();
      }

      # delete <meta>
#      elsif ($child_node->tag eq 'meta') {
#	$child_node->delete();
#      }

      # delete <script>
      elsif ($child_node->tag eq 'script') {
	$child_node->delete();
      }

      # delete <img> without text or links
      elsif ($child_node->tag eq 'img'
	     && !defined $child_node->attr('usemap')
	     && (!defined $child_node->attr('alt')
		 || $child_node->attr('alt') eq '')) {
	$child_node->delete();
      }

      # delete <font> tag
      elsif ($child_node->tag eq 'font') {
	foreach my $grandson ($child_node->content_refs_list) {
	  $child_node->preinsert($$grandson);
	  rule_delete_invisible_element($$grandson) if (ref $$grandson);
	}
	$child_node->delete();
      }

      # delete blank <tr>
      elsif ($child_node->tag eq 'tr') {
	my $blank = 1;
	foreach my $node_td ($child_node->content_refs_list) {
	  my $size = $$node_td->content_refs_list;
	  if ($size > 0) {
	    $blank = 0;
	    last;
	  }
	}
	if ($blank) {
	  $child_node->delete();
	} else {
	  rule_delete_invisible($child_node);
	}
      }

      # recursively check the child nodes
      else {
	rule_delete_invisible($child_node);
      }
}

# ----- rule_trans_invoicable.pl ----- #
sub rule_trans_invoicable($);
sub trans_map($);

sub rule_trans_invoicable($) {
  my $node = shift;
  bless $node, "HTML::Element";

  my @child_nodes = @{$node->content_array_ref};
  foreach my $child_node (@child_nodes) {
    if (ref $child_node) {
      if ($child_node->tag eq "img" && $child_node->attr('alt')) {
	$child_node->replace_with($child_node->attr('alt'))->delete();
      }

      # trans <br> to "\n"
      elsif ($child_node->tag eq 'br') {
#	$child_node->replace_with("\n")->delete();
      }

      # recursively check the child nodes
      else {
	rule_trans_invoicable($child_node);
      }
    }
  }
}

sub trans_map($) {
# input a map node
# output a reference of an array which contain links
  my $node_map = shift;
  bless $node_map, "HTML::Element";

  my @links;

  my @nodes = @{$node_map->content_array_ref};
  foreach my $node_area (@nodes) {
    if (ref $node_area) {
      if ($node_area->tag eq "area") {
	my $link = HTML::Element->new('a', 'href'=>$node_area->attr('href'), 'value'=>$node_area->attr('href'), 'class'=>$node_map->attr('name'));
	push(@links, $link);
      } else {
	warn "something wrong!";
      }
    } else {
      warn "something wrong!";
    }
  }

  # substitude img including usemap
  my $root = $node_map->root;
  my @node_imgs = $root->find_by_attribute('usemap', '#' . $node_map->attr('name'));
  foreach my $node_img (@node_imgs) {
    if ($node_img->tag eq 'img') {
	foreach my $link (@links) {
	  $node_img->preinsert($link);
	}
	$node_img->delete();
    } else {
      warn 'encount a usemap in <' . $node_img->tag . '>!';
    }
  }

  $node_map->delete();
}

# ----- rule_encoding.pl ----- #
sub rule_encoding($);

sub rule_encoding($) {
  my $node = shift;

  bless $node, "HTML::Element";

  my @child_nodes = $node->detach_content();
  foreach my $child_node (@child_nodes) {
    if (ref $child_node) {
      $node->push_content($child_node);
      rule_encoding($child_node);
    } else {
      # convert &nbsp from 0xA0 to ' '
      $child_node =~ tr/\xA0/ /;

      # convert to UTF-8

      # delete unknown characters
#      $child_node =~ tr/\x80-\xff//d;

      $node->push_content($child_node);
    }
  }
}

# ----- rule_restruct.pl ----- #
#require 'table.pl';

sub rule_restruct($);
sub rule_restruct_tree($);
sub rule_restruct_element($);
sub is_text_element($);
sub contain_only_text($);
sub wrap_text($);

sub rule_restruct($) {
  my $tree = shift;
  my $need_restart;

  # wrap all text. From now on, no bare text any more
  wrap_text($tree);

  rule_restruct_element($tree);
}

# deprecated!!!
sub rule_restruct_tree($) {
  my $node = shift;
#  bless $node, "HTML::Element";

  foreach my $child_node ($node->content_refs_list) {
    if (!($$child_node->tag eq 'div'
	  && ($$child_node->attr('class') eq 'e-guidedog_paragraph'
	      || $$child_node->attr('class') eq 'e-guidedog_text'
	      || $$child_node->attr('class') eq 'e-guidedog_table'))) {
      rule_restruct_element($$child_node);
    }
  }
}

sub rule_restruct_element($) {
  my $node = shift;
  bless $node, 'HTML::Element';

  # <li type='e-guidedog_paragraph'>
  if ($node->tag eq 'li' && contain_only_text($node)) {
    $node->attr('type', 'e-guidedog_paragraph');
    return;
  }

  # combine <b> <i> <em> <strong> <cite> <dfn> <code> <samp> <kbd> <var> <abbr> <acronym> <ins> <del> and <div class='e-guidedog_text>
  my @child_nodes = $node->detach_content();
  for (my $i = 0; $i <= $#child_nodes; $i++) {
    if (is_text_element($child_nodes[$i])) {
      if ($i < $#child_nodes && is_text_element($child_nodes[$i + 1])) {
	my $div = HTML::Element->new('div', 'class'=>'e-guidedog_paragraph');
	$div->push_content($child_nodes[$i++]);
	$div->push_content($child_nodes[$i]);
	while ($i <= $#child_nodes && is_text_element($child_nodes[$i + 1])) {
	  $div->push_content($child_nodes[$i++]);
	}
	$node->push_content($div);
      } elsif (ref $child_nodes[$i]) {
	my $div = HTML::Element->new('div', 'class'=>'e-guidedog_paragraph');
	$div->push_content($child_nodes[$i]);
	$node->push_content($div);
      } else {
	$node->push_content($child_nodes[$i]);
      }
    } else {
      if (!($child_nodes[$i]->tag eq 'div'
	    && ($child_nodes[$i]->attr('class') eq 'e-guidedog_paragraph'
		|| $child_nodes[$i]->attr('class') eq 'e-guidedog_table'))) {
	rule_restruct_element($child_nodes[$i]);
      }
      $node->push_content($child_nodes[$i]);
    }
  }
}


sub is_text_element($) {
  my $child_node = shift;
  if (!ref $child_node) {
    return 1;
  } else {
    bless $child_node, 'HTML::Element';
  }

  if ($child_node->tag eq 'b'
      || $child_node->tag eq 'i'
      || $child_node->tag eq 'tt'
      || $child_node->tag eq 'big'
      || $child_node->tag eq 'small'
      || $child_node->tag eq 'ins'
      || $child_node->tag eq 'del'
      || $child_node->tag eq 'em'
      || $child_node->tag eq 'strong'
      || $child_node->tag eq 'cite'
      || $child_node->tag eq 'dfn'
      || $child_node->tag eq 'code'
      || $child_node->tag eq 'samp'
      || $child_node->tag eq 'kdb'
      || $child_node->tag eq 'var'
      || $child_node->tag eq 'abbr'
      || $child_node->tag eq 'acronym'
#      || $child_node->tag eq 'ul'
#      || $child_node->tag eq 'ol'
      || ($child_node->tag eq 'div'
	  && ($child_node->attr('class') eq 'e-guidedog_text'
	      || $child_node->attr('class') eq 'e-guidedog_paragraph'))) {
    # test the children
    my $test = 1;
    foreach my $node ($child_node->content_refs_list) {
      if (!is_text_element($$node)) {
	$test = 0;
	last;
      }
    }
    return $test;
  }

  return 0;
}

sub contain_only_text($) {
  my $node = shift;
  if (!ref $node) {
    return 1;
  } else {
    bless $node, 'HTML::Element';
  }

  my $test = 1;
  foreach my $child_node ($node->content_refs_list) {
    if (!is_text_element($$child_node)) {
      $test = 0;
      last;
    }
  }

  return $test;
}

sub wrap_text($) {
  my $node = shift;

  my @children = $node->detach_content();
  foreach my $child (@children) {
    if (ref $child) {
      $node->push_content($child);
      wrap_text($child);
    } else {
     my $div = HTML::Element->new('div', 'class'=>'e-guidedog_text');
     $div->push_content($child);
     $node->push_content($div);
    }
  }
}

# ----- rule_struct_summary.pl ----- #
sub rule_struct_summary($);
sub struct_summary($$$);
sub rule_main_part($);
sub count_valid_text($);
sub count_valid_nodes($);

sub rule_struct_summary($) {
  my $tree = shift;

  my $text_len = count_valid_text($tree);
  my $total_nodes = count_valid_nodes($tree);
  my %levels;
  my %paras = (total_len => $text_len,
	      total_nodes => $total_nodes,
	      levels => \%levels);

  struct_summary($tree, 1, \%paras);

  ## recount levels

  # count max level
  my $max = 0;
  foreach (keys %levels) {
    $max = $_ if ($_ > $max);
  }

  # make a level mirror
  my $index = 1;
  for my $i (1 .. $max) {
    if ($levels{$i}) {
      $levels{$i} = $index++;
    }
  }

  # recount levels
  my @part_nodes = $tree->look_down('_tag', 'div',
				    'class', 'e-guidedog_part',
				    sub {defined $_[0]->attr('level')});
  my $parts = @part_nodes;
  my $i = 0;
  foreach my $part_node (@part_nodes) {
    $i++;
    $part_node->attr('level', $levels{$part_node->attr('level')});
    $part_node->attr('parts', $parts);
    $part_node->attr('part_no', $i);
  }
}

sub struct_summary($$$) {
  my $node = shift;
  my $level = shift;
  my $paras = shift;


  # 1) check whether it is a form
  if ($node->tag eq 'form') {
    my $div = HTML::Element->new('div',
				 'class' => 'e-guidedog_part',
				 'tags' => 'form',
				 'quantity' => 1,
				 'total' => 1,
				 'summary' => '',
				 'level' => $level);
    $paras->{levels}{$level} = 1; # mark this level has been used
    $node->replace_with($div);
    $div->push_content($node);
    goto level_finished;
  }

  # 2) check whether there is a list of same tags
  # count tags
  my $total = 0;
  my %tags;
  foreach my $child_node ($node->content_refs_list) {
#    if (is_sound_branch($$child_node) || (!is_silent_branch($$child_node) && !is_silent_node($$child_node))) {
    if (eGuideDog::Browser::HTML::Navigate::is_sound_branch($$child_node) || (!eGuideDog::Browser::HTML::Navigate::is_silent_branch($$child_node))) {
      $total++;
      if (ref $$child_node) {
	$tags{$$child_node->tag}++;
      } else {
	$tags{text}++;
      }
    }
  }
  my $min = $total * 0.8; # 0.8 is the check point, which can be changed
  if ($min >= 4) {
    foreach my $tag (keys %tags) {
      if ($tags{$tag} >= $min
	  && $tag ne 'label'
	  && $tag ne 'td'
	  && $tag ne 'area'
	  && $tag ne 'input'
	  && $tag ne 'link'
	  && $tag ne 'meta') {
	my $div = HTML::Element->new('div',
				     'class' => 'e-guidedog_part',
				     'tags' => $tag,
				     'quantity' => $tags{$tag},
				     'total' => $total,
				     'summary' => '',
				     'level' => $level);
	$paras->{levels}{$level} = 1; # mark this level has been used
	$node->replace_with($div);
	$div->push_content($node);
	goto level_finished;
      }
    }
  }

  # 3) check whether it has many text
  if ($node->tag ne 'body' # exclude some tags
      && $node->tag ne 'html') {
    my $len = count_valid_text($node);
    my $nodes_count = count_valid_nodes($node);
    if (($len >= $paras->{total_len} * 0.3 # 0.3 should be changed after experiment
	 || $len >= 1000)
	&& $nodes_count < $paras->{total_nodes} * 0.8) {
      # check childen' text length
      my $min_len = $len * 0.8; # should be changed after experiment
      my $has_main_child = 0;
      foreach my $child_node ($node->content_refs_list) {
	if (ref $$child_node) {
	  my $child_len = count_valid_text($$child_node);
	  if ($child_len > $min_len) {
	    $has_main_child = 1;
	    last;
	  }
	} else {
	  my $child_len = length($$child_node);
	  if ($child_len > $min_len) {
	  }
	}
      }

      # father become part if no main child
      if (!$has_main_child) {
	my $div = HTML::Element->new('div',
				     'class' => 'e-guidedog_part',
				     'tags' => $node->tag,
				     'quantity' => 1,
				     'length' => $len,
				     'total' => 1,
				     'summary' => '',
				     'level' => $level);
	$paras->{levels}{$level} = 1; # mark this level has been used
	$node->replace_with($div);
	$div->push_content($node);	
      }
    }
  }

  # recursively struct childen nodes
 level_finished:
  if ($node->tag ne 'form') { # form do not have child parts
    foreach my $child_node ($node->content_refs_list) {
      if (ref $$child_node
	  && !eGuideDog::Browser::HTML::Navigate::is_sound_branch($$child_node)
	  && !eGuideDog::Browser::HTML::Navigate::is_silent_branch($$child_node)) {
	struct_summary($$child_node, $level + 1, $paras);
      }
    }
  }
}

sub rule_main_part($) {
  my $tree = shift;

  my $text_len = count_valid_text($tree);

  # search engine type
  my $SEARCH_ENGINE_MAX_TEXT = 1024;
  if ($text_len < $SEARCH_ENGINE_MAX_TEXT) {
    my @forms = $tree->find_by_tag_name('form');
    if ($#forms == 0) {
      my @parts = $forms[0]->look_up('_tag', 'div',
				     'class', 'e-guidedog_part');
      $parts[0]->attr('mainpart', defined) if ($parts[0]);
    }
    return;
  }

  # main body with large text
  my @parts = $tree->look_down('_tag', 'div',
			       'class', 'e-guidedog_part');
  my $min_len = $text_len * 0.6; # 0.6 should be changed after experiment
  foreach my $part_node (@parts) {
    if (defined $part_node->attr('length')
	&& $part_node->attr('length') >= $min_len) {
      $part_node->attr('mainpart', defined);

      # clear all parents' mainpart tag
      my @parent_parts = $part_node->look_up('_tag', 'div',
					     'class', 'e-guidedog_part');
      # attation! 'look_up' include the node itself, so skip 0
      for my $i (1 .. $#parent_parts) {
	$parent_parts[$i]->attr('mainpart', undef);
      }
    }
  }
}

# count length of valid text of a tree node
sub count_valid_text($) {
  my $tree = shift;

  my $len = 0;

  $tree->traverse (
    [ # Callbacks;
     # pre-order callback:
     sub {
       my $node = $_[0];
       if (ref $node) {
	 if ($node->tag eq 'a'
	     || ($node->tag eq 'div'
		 && ($node->attr('class') eq 'e-guidedog_paragraph'
		     || $node->attr('class') eq 'e-guidedog_text'))) {
	   $len += length($node->as_text);
	   return 0;
	 } elsif ($node->tag eq 'select') { # option text are not counted
	   return 0;
	 }
       } else {
	 $len += length($node);
	 return 0;
       }
       return 1; # don't keep traversing
     },
     # post-order callback:
     undef
    ],
    0, # call the callbacks for text nodes
  );

  return $len;
}

sub count_valid_nodes($) {
  my $tree = shift;

  my $count = 0;

  $tree->traverse (
    [ # Callbacks;
     # pre-order callback:
     sub {
       my $node = $_[0];
       if (eGuideDog::Browser::HTML::Navigate::is_sound_branch($node)) {
	 $count++;
	 return 0;
       } elsif (eGuideDog::Browser::HTML::Navigate::is_silent_branch($node)) {
	 return 0;
       } elsif (!eGuideDog::Browser::HTML::Navigate::is_silent_node($node)) {
	 $count++;
	 return 1;
       }
       return 1; # keep traversing
     },
     # post-order callback:
     undef
    ],
    0, # call the callbacks for text nodes
  );

  return $count;
}

# ----- rule_add_number ----- #
sub rule_add_number($);

sub rule_add_number($) {
  my $tree = shift;

  my $no = 1;

  init_index($tree);

#  while (moveto_next_valid_node()) {
#    if (ref $index{current_node}) {
#      $index{current_node}->attr('e-guidedog_no', $no++);
#    } else {
#      $no++;
#    }
#  }
}

1;
